

#include <stdint.h>
#include <string.h>
#include "Timer.H"
#include "Lcd20x4.H"


using namespace avelino;
using namespace std;


const uint8_t Lcd20x4::STATUS_INIT_1      = 1;
const uint8_t Lcd20x4::STATUS_INIT_2      = 2;
const uint8_t Lcd20x4::STATUS_INIT_3      = 3;
const uint8_t Lcd20x4::STATUS_INIT_4      = 4;
const uint8_t Lcd20x4::STATUS_INIT_5      = 5;
const uint8_t Lcd20x4::STATUS_INIT_6      = 6;
const uint8_t Lcd20x4::STATUS_CHECK_QUEUE = 7;
const uint8_t Lcd20x4::STATUS_WAIT        = 8;


uint16_t Lcd20x4::internalWait() {
	uint16_t aux = 0;
	for (uint16_t i = 0; i < 50000; i++) {
		aux++;
	}
	return aux;
}


void Lcd20x4::init() {
	this->configurePorts();
	this->timerHandler = Timer::alloc();
	this->status = Lcd20x4::STATUS_INIT_1;
	uint8_t initCommands[5] = {
		0x28, 0x0C, 0x06, 0x01, 0x80
	};
	this->queueCommands(initCommands, 5);
}


void Lcd20x4::queueCommands(uint8_t *v, uint8_t n) {
	for (uint8_t i = 0; i < n; i++) {
		this->queue.push((uint16_t) (0x8000 | *v));
		v++;
	}
}


void Lcd20x4::queueSingleCommand(uint8_t v) {
	this->queue.push((uint16_t) (0x8000 | v));
}


void Lcd20x4::queueData(uint8_t *v, uint8_t n) {
	for (uint8_t i = 0; i < n; i++) {
		this->queue.push(*v);
		v++;
	}
}


void Lcd20x4::queueSingleData(uint8_t v) {
	this->queue.push(v);
}


void Lcd20x4::sendByte(uint8_t v) {
	this->setNibble((v >> 4) & 0x0F);
	this->setE(false);
	this->setE(true);
	this->setE(false);
	this->setNibble(v & 0x0F);
	this->setE(false);
	this->setE(true);
	this->setE(false);
}


void Lcd20x4::sendCommand(uint8_t v) {
	this->setRS(false);
	this->sendByte(v);
}


void Lcd20x4::sendData(uint8_t v) {
	this->setRS(true);
	this->sendByte(v);
}


void Lcd20x4::queueString(const char *s) {
	while (*s != 0) {
		this->queue.push((uint8_t) *s);
		s++;
	}
}


void Lcd20x4::queueLocate(uint8_t x, uint8_t y) {
	uint8_t address = 0;
	if (y == 1)
		address = 0x40;
	else if (y == 2)
		address = 0x14;
	else if (y == 3)
		address = 0x54;
	address += x;
	uint8_t v = 0x80 | (address & 0x7F);
	this->queueCommands(&v, 1);
}


void Lcd20x4::queueStringCenter(const char *s, uint8_t y) {
	uint8_t n = (uint8_t) strlen(s);
	this->queueLocate((n >> 1) - 10, y);
	this->queueString(s);
}


void Lcd20x4::loadBitmap(uint8_t c, uint8_t *bitmap) {
	this->queueSingleCommand(0x40 + (c << 3));
	this->queueData(bitmap, 8);
}


void Lcd20x4::run() {
	bool statusChanged;
	do {
		statusChanged = false;
		if (this->status == Lcd20x4::STATUS_INIT_1) {
			this->setRS(false);
			this->setE(false);
			this->status = Lcd20x4::STATUS_INIT_2;
			statusChanged = true;
			Timer::counter[this->timerHandler] = TIMER_MS(40);
		}
		else if (this->status == Lcd20x4::STATUS_INIT_2) {
			if (Timer::counter[this->timerHandler] == 0) {
				this->setNibble(0x03);
				this->setE(false);
				this->setE(true);
				this->setE(false);
				this->status = Lcd20x4::STATUS_INIT_3;
				statusChanged = true;
				Timer::counter[this->timerHandler] = TIMER_MS(40);
			}
		}
		else if (this->status == Lcd20x4::STATUS_INIT_3) {
			if (Timer::counter[this->timerHandler] == 0) {
				this->setNibble(0x03);
				this->setE(false);
				this->setE(true);
				this->setE(false);
				this->status = Lcd20x4::STATUS_INIT_4;
				statusChanged = true;
				Timer::counter[this->timerHandler] = TIMER_MS(40);
			}
		}
		else if (this->status == Lcd20x4::STATUS_INIT_4) {
			if (Timer::counter[this->timerHandler] == 0) {
				this->setNibble(0x03);
				this->setE(false);
				this->setE(true);
				this->setE(false);
				this->status = Lcd20x4::STATUS_INIT_5;
				statusChanged = true;
				Timer::counter[this->timerHandler] = TIMER_MS(40);
			}
		}
		else if (this->status == Lcd20x4::STATUS_INIT_5) {
			if (Timer::counter[this->timerHandler] == 0) {
				this->setNibble(0x02);
				this->setE(false);
				this->setE(true);
				this->setE(false);
				this->status = Lcd20x4::STATUS_INIT_6;
				statusChanged = true;
				Timer::counter[this->timerHandler] = TIMER_MS(40);
			}
		}
		else if (this->status == Lcd20x4::STATUS_INIT_6) {
			if (Timer::counter[this->timerHandler] == 0) {
				this->status = Lcd20x4::STATUS_CHECK_QUEUE;
				statusChanged = true;
			}
		}
		else if (this->status == Lcd20x4::STATUS_CHECK_QUEUE) {
			if (this->queue.hasData()) {
				uint16_t v = this->queue.pop();
				if ((v & 0x8000) != 0)
					this->sendCommand((uint8_t) (v & 0x00FF));
				else
					this->sendData((uint8_t) v);
				Timer::counter[this->timerHandler] = TIMER_MS(10);
				this->status = Lcd20x4::STATUS_WAIT;
				statusChanged = true;
			}
		}
		else if (this->status == Lcd20x4::STATUS_WAIT) {
			if (Timer::counter[this->timerHandler] == 0) {
				this->status = Lcd20x4::STATUS_CHECK_QUEUE;
				statusChanged = true;
			}
		}
	} while (statusChanged);
}
